// mtbw.c: S.W. Ellingson, Virginia Tech, 2010 Sep 30
// ---
// COMPILE: (see Makefile)
// ---
// COMMAND LINE SYNTAX, INPUT, OUTPUT: 
//   See help message (search for "Help Message follows:" in source code below)
// ---
// REQUIRES:
// mtbw.h 
// ---
// TBW data acquisition, parsing, analysis, and reformatting
// See end of this file for history.

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "mtbw.h"
#include "nr_gasdev.c"
#include "nr_ran1.c"

#define MY_NAME "mtbw (v.20100930.1)"
#define ME "16" /* listed in LWA_MCS.h */

/* --- eSource values ---*/
/* positive integers refer to DRs */
#define FD       -1
#define FC       -2  /* not yet implemented */
#define TEST_SIN -3
#define TEST_WGN -4

#define MAX_FILENAME_LENGTH 256
#define N_SAMPLES_SIM       12000000 /* number of samples in sim */
#define OMEGA            (6.2832e+7) /* = 2*pi*(10e+6) [rad/s] */
#define FS                (196.0e+6) /* [samples/s] */

/* This could be huge */
static short int x[TBWH_N_SAMPLES_MAX];

main ( int narg, char *argv[] ) {

  /*=================*/
  /*=== Variables ===*/
  /*=================*/
 
  int iarg;

  int eSource = TEST_SIN;                 /* <source>; set by "-s" switch */
  char src_filename[MAX_FILENAME_LENGTH]; /* <filename>; set by "-sFD=" switch */
  int nSensor=0;                          /* <sensor>; set by "-n" switch */
  int bVerbose=0;                         /* set by "-v" switch */

  int bDone;
  int bDone2;

  int eStatus = -99;    /* return code from tbw_file_info (in tbw.h) */
  int b4bit = -99;      /* see tbw_file_info (in tbw.h) */
  long int nFrames = -99;   /* see tbw_file_info (in tbw.h) */
  int iStandMax = -99;  /* see tbw_file_info (in tbw.h) */
  long int time0 = -99;  /* see tbw_file_info (in tbw.h) */
  
  int i;
  int iSensor1, iSensor2;
  int iStand;
  int iPol;

  long int k,l;

  long int lseed;

  short int min, max;
  short int min_code, max_code;
  float mean, std;
  long int ccp, ccm;

  FILE *fp1;
  FILE *fp2;
  char filename[256];

  /*==================*/
  /*=== Initialize ===*/
  /*==================*/
    
  /* First, announce thyself */
  printf("[%s] I am %s\n",ME,MY_NAME);

  /* clearing this just in case */
  sprintf(src_filename,"");

  /* seed for the random number generator used for TEST_WGN */
  lseed = (long int) time(NULL); 

  /* process command line arguments */
  for ( iarg=2; iarg<=narg; iarg++) {

    bDone2 = 0;

    /* sense & process "-h" */
    if (!strncmp(argv[iarg-1],"-h",2)) {
      bDone2 = 1;
      /* Help Message follows: */
      printf("COMMAND LINE:\n");
      printf("  mtbw -h -s<source> -n<sensor> -v -c\n");
      printf("WHAT I DO:\n");
      printf("  Acquire, read, or simulate data from a single TBW trigger.\n");
      printf("  Parse data into into separate files per sensor, plus a file for time tags.\n");
      printf("  Compute diagnostic statistics from sensor data.\n");
      printf("  Pre-processor for the data visualization program mtbwv.\n");
      printf("INPUT:\n");
      printf("  -h means show help, then exit.\n");
      printf("  -s specifies source of data to analyze.  Options for <source> are:\n");
      printf("     DR#; e.g., DR1 means get a fresh acquisition via DR1.\n");
      printf("       (Must be the one connected to TBW!)\n"); 
      printf("     FD=<filename> means use <filename>; assumed to be from a DR.\n");
      //printf("     FC=<filename> means use <filename>, assumed to be single-\n");
      //printf("       sensor space-delimited ASCII data\n");
      printf("     test_sin means use a single-sensor simple internally-generated test\n");
      printf("       sinusoid. 12,000,000 samples of 12-bit data. [DEFAULT]\n");
      printf("     test_wgn means use a single-sensor simple internally-generated white\n");
      printf("       gaussian noise. 12,000,000 samples of 12-bit data.\n");
      printf("  -n specifies sensor; meaningful only if <source> is DR or FD.\n");
      printf("     <sensor> is in the range 1 - 520 (less if TBW supports less).\n");
      printf("     1 & 2 correspond to Stand 1; 3 & 4 correspond to Stand 2; and so on\n");
      printf("     (This is DP-output numbering; may not correspond to actual stand/pol IDs)\n");
      printf("     <sensor>=0 means process all available sensors [DEFAULT]\n");
      printf("  -v invokes verbose output. Default is not to be verbose.\n");
      printf("  -c (clean) deletes out files from previous run(s) and exits. Default is don't.\n");
      printf("COMMAND LINE EXAMPLES:\n");
      printf("  ./mtbw -sFD=tbw_data_07_28_19.7764\n");
      printf("    Reads local file tbw_data_07_28_19.7764, parses it, analyzes all sensors\n");
      printf("  ./mtbw -sFD=tbw_data_07_28_19.7764 -n1\n");
      printf("    Reads local file tbw_data_07_28_19.7764, parses it, analyzes sensor #1\n");
      printf("OUTPUT:\n");
      printf("  mtbw###.dat (where ###=<sensor>): time series (A/D codes), binary format.\n");
      printf("    Intended primary for mtbwv, but format is simple (see mtbw.h)\n");
      printf("  mtbw_tt.dat: Table of time tags seen.\n");
      printf("    Columns=stands, rows=frames, decimal in space-delimited ASCII format.\n");
      printf("    Shown are the actual time tags minus the value of first time tag seen.\n");
      printf("    (Value of first time is written as a status message to stdout.)\n");
      printf("  mtbw_stats.dat: Statistics for <sensor> (or all sensors, depending on -n).\n");
      printf("    ASCII, space-delimited, one sensor per line.  Line format follows:\n");
      printf("    sensor_ID mean std_dev min_code max_code neg_clips pos_clips\n");
      printf("      min/max_code are smallest/largest A/D codes seen.\n");
      printf("      neg/pos_clips are numbers of 0's/4095's seen.\n");
      printf("  mtbw_t###.dat: Time series for sensor ###, in A/D codes, ASCII.\n");
      return;
      } /* if (!strncmp(argv[iarg-1],"-FD=",4))  */

    /* sense & process "-s" */
    //printf("%d\n",strncmp(argv[1],"-s",2));
    if (!strncmp(argv[iarg-1],"-s",2)) {
      bDone2 = 1;
      bDone = 0;
      //printf("<%s>\n",argv[iarg-1]);
      if (!strncmp(argv[iarg-1],"-sDR",4)) {
        bDone = 1;
        if (strlen(argv[iarg-1])<5) {
          printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
          return;
          }
        sscanf( &(argv[iarg-1][4]), "%d", &eSource ); 
        //printf("DR <%s> <%d>\n",&(argv[iarg-1][4]),eSource); 
        }         
      if (!strncmp(argv[iarg-1],"-sFD",4)) {
        bDone = 1;
        //printf("FD\n");
        eSource = FD; 
        if (strlen(argv[iarg-1])<6) {
          printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
          return;
          }
        sscanf( &(argv[iarg-1][5]), "%s", src_filename ); 
        } 
      //if (!strncmp(argv[iarg-1],"-sFC=",4)) {
      //  bDone = 1;
      //  //printf("FC\n"); 
      //  eSource = FC;
      //  if (strlen(argv[iarg-1])<6) {
      //    printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
      //    return;
      //    }
      //  sscanf( &(argv[iarg-1][5]), "%s", src_filename ); 
      //  } 
      if (!strncmp(argv[iarg-1],"-stest_sin",10)) {
        bDone = 1;
        //printf("test_sin\n");
        eSource = TEST_SIN; 
        } 
      if (!strncmp(argv[iarg-1],"-stest_wgn",10)) {
        bDone = 1;
        // printf("test_wgn\n"); 
        eSource = TEST_WGN; 
        }
      if (!bDone) {
        printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
        return;
        }
      } /* if (!strncmp(argv[iarg-1],"-s",2)) */

    /* sense & process "-n" */
    if (!strncmp(argv[iarg-1],"-n",2)) {
      bDone2 = 1;
      if (strlen(argv[iarg-1])<3) {
        printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
        return;
        }
      sscanf( &(argv[iarg-1][2]), "%d", &nSensor );       
      } /* if (!strncmp(argv[iarg-1],"-n",2))  */

    /* sense & process "-v" */
    if (!strncmp(argv[iarg-1],"-v",2)) {
      bDone2 = 1;
      bVerbose = 1;
      } /* if (!strncmp(argv[iarg-1],"-v",2))  */

    /* sense & process "-c" */
    if (!strncmp(argv[iarg-1],"-c",2)) {
      bDone2 = 1;
      system("rm mtbw???.dat"); /* takes care of mtbw_tt.dat as well */
      system("rm mtbw_stats.dat");
      system("rm mtbw_t???.dat");
      return;
      } /* if (!strncmp(argv[iarg-1],"-c",2))  */

    if (!bDone2) {
      printf("[%s] FATAL: Can't parse switch <%s>\n",ME,argv[iarg-1]);
      return;
      }
 
    } /* for (iarg */

  /* Show what we know about this run... */
  printf("[%s] PARAMETER: eSource = %d\n",ME,eSource);
  printf("[%s] PARAMETER: src_filename = <%s>\n",ME,src_filename); 
  printf("[%s] PARAMETER: nSensor = %d\n",ME,nSensor); 
  printf("[%s] PARAMETER: bVerbose = %d\n",ME,bVerbose); 

  if (eSource>0) { /* source is a data recorder.  need to acquire a file. */
    printf("[%s] FATAL: -sDR# option disabled.\n",ME);  
    return; 
    }
  /* src_filename should now be filename of acquired file */

  if ( (eSource>0) || (eSource==FD) ) { /* source is a previously-acquired file. */

    printf("[%s] Stand by: reading <%s>...\n",ME,src_filename);
    eStatus = tbw_file_info(src_filename, &b4bit, &nFrames, &iStandMax, &time0, bVerbose );
    /* at this point tbw###.dat files should exist locally */

    printf("[%s] eStatus = %d (0 or -2 are OK; see mtbw.h for others)\n",ME,eStatus);
    printf("[%s] b4bit = %d (1 means 4-bit data, 0 means 12-bit data; -99 if not seen)\n",ME,b4bit);
    printf("[%s] nFrames = %ld (number of 0xDEC0DE5C's seen)\n",ME,nFrames);
    printf("[%s] iStandMax = %d (highest stand index seen (1..520; 0 if not seen))\n",ME,iStandMax);
    printf("[%s] time0 = %ld (first time tag seen; -99 if not seen)\n",ME,time0);
    //printf("[%s] %ld snapshots found, corresponding to %f ms\n",ME,(nFrames/iStandMax)*(400+b4bit*800),
    //                                                              ((nFrames/iStandMax)*(400+b4bit*800))/(196.0e+3));
    if ( (eStatus==-1) || (eStatus<-2) ) {
      printf("[%s] FATAL: Problem reading <%s>. Check eStatus (above) for more info.\n",ME,src_filename);  
      return; 
      }

    } /* if ( (eSource>0) || (eSource==FD) ) */

  /* Figure out range of sensor IDs to be considered */
  iSensor1 = nSensor; /* if specific sensor is specified, we do just that sensor */
  iSensor2 = nSensor; 
  if (nSensor==0) {   /* if all sensors are to be done: */
    iSensor1 = 1;
    iSensor2 = 2*iStandMax;
    }
  if (eSource<-2) {  /* then this is simulation (TEST_SIN or TEST_WGN) */
    iSensor1 = 1; /* this is just a kludge to get into the main processing loop below */
    iSensor2 = 1;
    }

  /* Create mock data and write it to mtbw001.dat */
  if ( (eSource==TEST_SIN) || (eSource==TEST_WGN) ) {
    fp1 = fopen("mtbw001.dat","wb");
    for (l=0;l<N_SAMPLES_SIM;l++) {
      if (eSource==TEST_SIN) {
        x[l] = (short int) ( 2048 + 2000*sin(OMEGA*l/FS) );
        }
      if (eSource==TEST_WGN) {
        //x[l] = (short int) ( 48 + ( rand() % 4000 ) );
        x[l] = (short int) ( 2048 + 500*gasdev(&lseed) );
        }
      fwrite(&(x[l]),sizeof(short int),1,fp1);
      fwrite(&(x[l]),sizeof(short int),1,fp1);  /* just copy same stuff to other pol */
      }
    fclose(fp1);
    b4bit=0; /* has to be set so that clip counter works correctly */
    printf("[%s] Made %ld samples\n",ME,l);
    }

  fp1 = fopen("mtbw_stats.dat","w");

  /* Loop through sensors */
  for ( i=iSensor1; i<=iSensor2; i++ ) {
    //printf("[%s] Doing Sensor=%d\n",ME,i);

    /* For each sensor considered, we need to get data into memory */
    /* How we do this depends on where the data is currently at */

    if ( (eSource>0) || (eSource==FD) ) { /* source is a previously-acquired file. */ 
      iStand = (i-1)/2 + 1;
      iPol = i-2*(iStand-1); 
      printf("[%s] Now looking at iStand=%d, iPol=%d\n",ME,iStand,iPol);  
      if (!(tbw_sfile_import( iStand, iPol, b4bit, x, &l ))) {
        printf("[%s] FATAL: tbw_sfile_import(%d) failed.\n",ME,i);
        return; 
        }
      printf("[%s] Found %ld samples\n",ME,l);
      } /* if ( (eSource>0) || (eSource==FD) ) */

    /* At this point, data is in memory regardless of source/format */

    /* Compute first-pass statistics */
    mean = 0;   /* mean */
    min = x[0]; /* min value */
    max = x[0]; /* max value */
    ccp = 0;    /* positive clip count */
    ccm = 0;    /* negative clip count */
    min_code = 0;
    max_code = 4095;
      if (b4bit) max_code = 15; 
    for (k=0;k<l;k++) {
      mean += ((float)x[k])/l;
      if (x[k]<min) { min=x[k]; }
      if (x[k]>max) { max=x[k]; }
      if (x[k]>=max_code) { ccp++; /* printf("%d %d\n",x[k],max_code); */ }
      if (x[k]<=min_code) { ccm++; } 
      }

    /* Compute second-pass statistics */
    std = 0;   /* standard deviation */
    for (k=0;k<l;k++) {
      std += ( x[k] - mean ) * ( x[k] - mean ) / l; /* interim; taking sqrt later */
      }
    std = sqrt(std);

    printf("[%s] %d %f %f %d %d %ld %ld\n",ME,i,mean,std,min,max,ccm,ccp);
    fprintf(fp1,"%03d %6.1f %6.1f %4d %4d %10ld %10ld\n",   i,mean,std,min,max,ccm,ccp);

    /* Write a copy of the time-domain waveform */
    sprintf(filename,"mtbw_t%03d.dat",i);
    fp2 = fopen(filename,"w");
    for (k=0;k<l;k++) fprintf(fp2,"%04d\n",x[k]);
    fclose(fp2);

    } /* for ( i=iSensor1 ... */

  fclose(fp1);

  return;
  } /* main() */

//==================================================================================
//=== HISTORY ======================================================================
//==================================================================================
// mtbw.c: S.W. Ellingson, Virginia Tech, 2010 Sep 23
//   .1: initial version
// mtbw.c: S.W. Ellingson, Virginia Tech, 2010 Sep 30
//   .1: removed diagnostic printf that shows clipping instances 
//       added -v (verbose) option
//       added -c (clean: remove output files from previous run) option
//       help message discloses the mtbw_t???.dat files

//==================================================================================
//=== BELOW THIS LINE IS SCRATCH ===================================================
//==================================================================================

