// mtbwv.c: S.W. Ellingson, Virginia Tech, 2010 Sep 30
// ---
// COMPILE: See Makefile
// ---
// COMMAND LINE SYNTAX, INPUT, OUTPUT: 
//   See help message (search for "Help Message follows:" in source code below)
// ---
// REQUIRES:
// gnuplot_i (http://ndevilla.free.fr/gnuplot/gnuplot_i/index.html)
// FFTW
// ---
// TBW data visualization.  Requires output from mtbw
// See end of this file for history.

#include <stdio.h>
#include <math.h>
#include <fftw3.h>
#include "gnuplot_i.h"

#define MY_NAME "mtbwv (v.20100930.1)"
#define ME "17" /* listed in LWA_MCS.h */

#define TBWH_NANOSEC_PER_SAMPLE   5.1020     /* ns/sample at 196 MSPS */
#define TBWH_MILLISEC_PER_SAMPLE (5.1020e-6) /* ms/sample at 196 MSPS */
#define TBWH_N_SAMPLES_MAX 36000000
#define TBWH_FS (196.0e+6)                   /* TBW samples/s */
#define TBWH_NSB_MAX 262144

static short int tx[TBWH_N_SAMPLES_MAX]; /* data read from mtbw###.dat (could be huge) */
static float     ax[TBWH_N_SAMPLES_MAX];     /* time-averaged data */

static float S[TBWH_NSB_MAX/2]; /* last spectrum */
static float P[TBWH_NSB_MAX/2]; /* integrated spectrum */

int main( int narg, char *argv[] ) {

  int iarg;
  int bDone;

  int n=1;              /* sensor to look at */
  int b4bit=0;          /* "1" means 4-bit data; "0" means 12-bit data */
  long int iBegin=1;    /* sample to begin at */ 
  long int nLength=301; /* number of samples. 0 means "to end of file" */
  int bDoWaveform=1;    /* =1 means show time-domain waveform */
  int bDoAvePower=0;    /* =1 means show time-averaged waveform */
  int bDoAveSpectrum=0; /* =1 means show averaged spectrum */
  long int nsb=1024;    /* number of samples per block = FFT length = time-domain averaging length */
  int bRemoveBias=0;    /* =1 calculate and remove zero-freq bias, block-by-block */
  int bdB=0;            /* =1 plot spectrum in dB */

  FILE *fp;
  char filename[256];

  int iStand;
  int iPol;

  long int k;
  short int x12, y12;   /* used to read 12-bit data */
  char c;               /* used to read 4-bit data */

  long int u;
  long int q;
  long int r;

  float sum;

  /* FFTW variables */
  fftw_complex *in, *out;
  fftw_plan p;

  long int l;

  /* Gnuplot variables */
  gnuplot_ctrl *h1;
  gnuplot_ctrl *h2;
  gnuplot_ctrl *h3;

  /*==================*/
  /*=== Initialize ===*/
  /*==================*/
    
  /* First, announce thyself */
  printf("[%s] I am %s\n",ME,MY_NAME);
  printf("[%s] This software uses the gnuplot_i library written by N. Devillard\n",ME);

  /* process command line arguments */
  for ( iarg=2; iarg<=narg; iarg++) {

    bDone = 0;

    /* sense & process "-h" */
    if (!strncmp(argv[iarg-1],"-h",2)) {
      bDone = 1;
      /* Help Message follows: */
      printf("COMMAND LINE:\n");
      printf("  mtbwv -h -n<sensor> -4 -b<sample> -l<samples> -x -a -s -t<nsb> -o -d\n");
      printf("INPUT:\n");
      printf("  -h means show help, then exit.\n");
      printf("  -n<sensor> means look at <sensor> (DEFAULT: -n1).\n");
      printf("    1 & 2 correspond to Stand 1; 3 & 4 correspond to Stand 2; and so on\n");
      printf("  -4 means assume 4-bit data (DEFAULT: assume 12-bit data).\n");
      printf("  -b<sample> means begin at <sample> (DEFAULT: -b1).\n");
      printf("  -l<samples> means use <samples> samples (DEFAULT: -l301).\n");
      printf("  -x means DON'T show time-domain waveform (DEFAULT: show it).\n");
      printf("  -a means show averaged power (DEFAULT: don't).\n");
      printf("  -s means show averaged spectrum (DEFAULT: don't).\n");
      printf("  -t<nsb> means use <samples> samples per block (DEFAULT: -t1024).\n");
      printf("    This is averaging time for averaged-power output, and\n");
      printf("    FFT length for averaged-spectrum output\n");
      printf("  -o means remove zero-frequency bias before FFT (DEFAULT: don't).\n");
      printf("  -d means plot spectrum in dB (DEFAULT: linear units).\n");
      printf(" OUTPUT FILES:\n");
      printf("   mtbwv_t.dat (waveform data), unless -x is used.\n");
      printf("   mtbwv_a.dat (averaged power), if -a is used.\n");
      printf("   mtbwv_s.dat (averaged spectrum), if -s is used.\n");
      return;
      } 

    /* sense & process "-n" */
    if (!strncmp(argv[iarg-1],"-n",2)) {
      bDone = 1; 
      if (strlen(argv[iarg-1])<3) {
        printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
        return;
        }
      sscanf( &(argv[iarg-1][2]), "%d", &n ); 
      if ( (n<1) || (n>520) ) {
        printf("[%s] FATAL: Invalid sensor index: %d\n",ME,n);
        return;
        }
      } 

    /* sense & process "-4" */
    if (!strncmp(argv[iarg-1],"-4",2)) {
      bDone = 1;
      b4bit = 1;
      }

    /* sense & process "-b" */
    if (!strncmp(argv[iarg-1],"-b",2)) {
      bDone = 1;   
      if (strlen(argv[iarg-1])<3) {
        printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
        return;
        }
      sscanf( &(argv[iarg-1][2]), "%ld", &iBegin ); 
      } 

    /* sense & process "-l" */
    if (!strncmp(argv[iarg-1],"-l",2)) {
      bDone = 1;   
      if (strlen(argv[iarg-1])<3) {
        printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
        return;
        }
      sscanf( &(argv[iarg-1][2]), "%ld", &nLength ); 
      } 

    /* sense & process "-x" */
    if (!strncmp(argv[iarg-1],"-x",2)) {
      bDone = 1;
      bDoWaveform = 0;
      }

    /* sense & process "-a" */
    if (!strncmp(argv[iarg-1],"-a",2)) {
      bDone = 1;
      bDoAvePower = 1;
      }

    /* sense & process "-s" */
    if (!strncmp(argv[iarg-1],"-s",2)) {
      bDone = 1;
      bDoAveSpectrum = 1;
      }

    /* sense & process "-t" */
    if (!strncmp(argv[iarg-1],"-t",2)) {
      bDone = 1;   
      if (strlen(argv[iarg-1])<3) {
        printf("[%s] FATAL: Can't parse parameter of recognized switch <%s>\n",ME,argv[iarg-1]);
        return;
        }
      sscanf( &(argv[iarg-1][2]), "%ld", &nsb ); 
      }

    /* sense & process "-o" */
    if (!strncmp(argv[iarg-1],"-o",2)) {
      bDone = 1;
      bRemoveBias = 1;
      }

    /* sense & process "-d" */
    if (!strncmp(argv[iarg-1],"-d",2)) {
      bDone = 1;
      bdB = 1;
      }

    if (!bDone) {
      printf("[%s] FATAL: Can't parse switch <%s>\n",ME,argv[iarg-1]);
      return;
      }
 
    } /* for (iarg */

  printf("[%s] PARAMETER: n = %d\n",ME,n);
  printf("[%s] PARAMETER: b4bit = %d\n",ME,b4bit);
  printf("[%s] PARAMETER: iBegin = %ld\n",ME,iBegin);
  printf("[%s] PARAMETER: nLength = %ld\n",ME,nLength);
  printf("[%s] PARAMETER: bDoWaveform = %d\n",ME,bDoWaveform);
  printf("[%s] PARAMETER: bDoAvePower = %d\n",ME,bDoAvePower);
  printf("[%s] PARAMETER: bDoAveSpectrum = %d\n",ME,bDoAveSpectrum);
  printf("[%s] PARAMETER: nsb = %ld\n",ME,nsb);
  printf("[%s] PARAMETER: bRemoveBias = %d\n",ME,bRemoveBias);
  printf("[%s] PARAMETER: bdB = %d (1=plot in dB; 0=plot in linear units)\n",ME,bdB);
  //return;

  if ( ( bDoAvePower || bDoAveSpectrum) && ( nsb>nLength ) ) {
    printf("[%s] FATAL: nsb > nLength not allowed if -a or -s are set.\n",ME);
    return;
    }
  if ( nsb>TBWH_NSB_MAX ) {
    printf("[%s] FATAL: nsb > %d not allowed.\n",ME,TBWH_NSB_MAX);
    return;
    }

  /* Identify stand (so we know which mtbw###.dat file to use) and pol */
  iStand = (n+1)/2;
  iPol = n - 2*(iStand-1);
  //printf("[%s] %d %d %d\n",ME,n,iStand,iPol);
  //return;

  /* === Getting time domain data from file === */
  /* Data ends up in tx[], which goes 0..k */

  sprintf(filename,"mtbw%03d.dat",iStand);
  if (!(fp = fopen(filename,"rb"))) { 
    printf("[%s] FATAL: Can't open <%s>.  (Did you forget to use mtbw first?)\n",ME,filename); 
    return;
    } 

  k = 0;
  if (!b4bit) { /* 12 bit data */
      while ( fread(&x12,sizeof(short int),1,fp) > 0 ) {
        fread(&y12,sizeof(short int),1,fp);
        k++;
        if (iPol==1) { tx[k-1] = x12; }
        if (iPol==2) { tx[k-1] = y12; }
        }
    } else { /* 4 bit data */
      while ( fread(&c,sizeof(char),1,fp) > 0 ) {
        k++;
        if (iPol==1) { tx[k-1] = (short int) ( 0x000F & (( c >> 4   ) ^ 0x08 ) ); }
        if (iPol==2) { tx[k-1] = (short int) ( 0x000F & (( c & 0x0F ) ^ 0x08 ) ); }  
        }
    }

  printf("[%s] Read %ld samples from %s.\n",ME,k,filename);
  
  fclose(fp);

  /* ========================================== */
  /* === Do display of time domain waveform === */
  /* ========================================== */

  if (bDoWaveform) {

    /* Writing back out the dataset we want to show */
    fp = fopen("mtbwv_t.dat","w");
    for ( k = (iBegin-1) ; k < (iBegin-1+nLength+1) ; k++ ) {
      fprintf(fp,"%10.1f %5d\n", ((float)k)*TBWH_NANOSEC_PER_SAMPLE, tx[k] );
      }
    fclose(fp);    
  
    /* Time domain plot */

    h1 = gnuplot_init();
    gnuplot_cmd(h1,"plot 'mtbwv_t.dat' using 1:2 with linespoints");
    if (b4bit) {
        gnuplot_cmd(h1,"set yrange [0:15]");
      } else {
        gnuplot_cmd(h1,"set yrange [0:4095]");
      }  
    gnuplot_cmd(h1,"set xlabel 'Time [ns]'");
    gnuplot_cmd(h1,"set ylabel 'A/D Code'"); 
    gnuplot_cmd(h1,"set grid; show grid");  
    gnuplot_cmd(h1,"replot");

    } /* if (bDoWaveform) */

  /* ============================================ */
  /* === Do display of average power waveform === */
  /* ============================================ */

  if (bDoAvePower) {

    /* Doing the time averaging: */
    k = iBegin; /* sample index */
    u = 0;      /* counting samples considered */
    q = 0;      /* counting samples in a block */
    r = 0;      /* counting blocks */
    ax[r] = 0;
    while ( (u<nLength) && (k<TBWH_N_SAMPLES_MAX) ) {
      ax[r] += ((float)tx[k])/nsb;
      k++;
      u++;
      q++; 
      if (q==nsb) {
        q = 0; /* reset counting of samples within a block */
        r++;   /* starting new block */
        ax[r] = 0;
        } /* if (q==nsb) */      
      } /* while (p<nLength) */
    /* r is now taken to be number number of blocks.  Technically we have */
    /* one more than this, but the last block potentially represents < nsb samples */

    /* Writing back out the dataset we want to show */
    fp = fopen("mtbwv_a.dat","w");
    for ( k = 0 ; k < r ; k++ ) {
      fprintf(fp,"%10.7f %7.1f\n", ((float)k)*nsb*TBWH_MILLISEC_PER_SAMPLE, ax[k] );
      }
    fclose(fp);    
  
    /* Average power (over time) plot */

    h2 = gnuplot_init();
    gnuplot_cmd(h2,"plot 'mtbwv_a.dat' using 1:2 with lines");
    if (b4bit) {
        gnuplot_cmd(h2,"set yrange [0:15]");
      } else {
        gnuplot_cmd(h2,"set yrange [0:4095]");
      }  
    gnuplot_cmd(h2,"set xlabel 'Time [ms]'");
    gnuplot_cmd(h2,"set ylabel 'Average [A/D Codes]'");
    gnuplot_cmd(h2,"set title 'Int. Time = %10.3e [s]'",((float)nsb)/TBWH_FS); 
    gnuplot_cmd(h2,"set grid; show grid");  
    gnuplot_cmd(h2,"replot");

    } /* if (bDoAvePower) */

  /* ============================================ */
  /* === Do display of averaged spectrum ======== */
  /* ============================================ */

  if (bDoAveSpectrum) {

    /* FFTW setting up */
    in = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsb);
    out = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * nsb);
    p = fftw_plan_dft_1d(nsb, in, out, FFTW_FORWARD, FFTW_MEASURE);

    /* initialize PSD array */
    for (q=0;q<(nsb/2);q++) P[q]=0.0;

    /* --- Main loop --- */
    k = iBegin; /* sample index */
    u = 0;      /* counting samples considered */
    r = 0;      /* counting blocks */
    ax[r] = 0;
    while ( (u<nLength) && ((k+nsb-1)<TBWH_N_SAMPLES_MAX) ) {
     
      /* load next nsb samples into FFT input array */
      q = 0;
      sum = 0; /* for calculating bias */
      while (q<nsb) {
        in[q][0] = (float) tx[k];
        in[q][1] = 0.0;
        out[q][0] = 0.0; /* clearing output array, just in case */
        out[q][1] = 0.0;
        k++;
        u++;
        q++; 
        sum += in[q][0];
        } 
         
      /* remove bias if requested */
      if (bRemoveBias) {
        sum /= nsb;
        for (q=0;q<nsb;q++) {
          in[q][0] -= sum;
          }
        }

      /* do that crazy FFT thing */	
      fftw_execute(p); 

      /* average output with previous FFT squared-magnitudes */
      /* calc S and P */
      for (q=0;q<(nsb/2);q++) {
        S[q] = ( ( out[q][0] * out[q][0] ) + ( out[q][1] * out[q][1] ) ) / nsb;
        P[q] += S[q];
        }    
  
      /* update number of blocks */
      r++; 

      } /* while ( (u<nLength) && ((k+nsb-1)<TBWH_N_SAMPLES_MAX) ) */

    /* shut down FFTW */
    fftw_destroy_plan(p);
    fftw_free(in); 
    fftw_free(out);

    /* complete calculation of average */
    for (q=0;q<(nsb/2);q++) P[q] /= r;

    /* Writing back out the dataset we want to show */
    fp = fopen("mtbwv_s.dat","w");
    for ( q = 0 ; q < (nsb/2) ; q++ ) {
      if (bdB) {
          fprintf(fp,"%10.6f %7.1f\n", (q*TBWH_FS/nsb)/(1.0e+6), 10.0*log10(P[q]) );
        } else {
          fprintf(fp,"%10.6f %7.1f\n", (q*TBWH_FS/nsb)/(1.0e+6), P[q] );
        }
      }
    fclose(fp);    
  
    /* plot */
    h3 = gnuplot_init();
    gnuplot_cmd(h3,"plot 'mtbwv_s.dat' using 1:2 with lines");
    gnuplot_cmd(h3,"set xrange [1:98]"); 
    //gnuplot_cmd(h3,"set yrange [0:4095]"); 
    gnuplot_cmd(h3,"set xlabel 'Freq [MHz]'");
    if (bdB) {
        gnuplot_cmd(h3,"set ylabel 'PSD [dB(/RBW)]'");
      } else { 
        gnuplot_cmd(h3,"set ylabel 'PSD [/RBW]'");
      }
    gnuplot_cmd(h3,"set title 'Int. Time = %10.3e [s]. RBW = %10.3e [Hz]'",((float)(nsb*r))/TBWH_FS,((float)TBWH_FS)/nsb);
    gnuplot_cmd(h3,"set grid; show grid");  
    gnuplot_cmd(h3,"replot");

    } /* if (bDoAvePower) */

  /* ========================================== */
  /* === Clean up ============================= */
  /* ========================================== */

  if ( bDoWaveform || bDoAvePower || bDoAveSpectrum ) {
    printf("Hit ENTER to continue..."); getchar();
    }
  if (bDoWaveform)    gnuplot_close(h1);
  if (bDoAvePower)    gnuplot_close(h2);
  if (bDoAveSpectrum) gnuplot_close(h3);

  return 0;
  } /* main() */

//==================================================================================
//=== HISTORY ======================================================================
//==================================================================================
// mtbwv.c: S.W. Ellingson, Virginia Tech, 2010 Sep 27
//   .1: initial version
// mtbwv.c: S.W. Ellingson, Virginia Tech, 2010 Sep 30
//   .1: adding -d switch

//==================================================================================
//=== BELOW THIS LINE IS SCRATCH ===================================================
//==================================================================================
